@@ -1,4 +1,4 @@
-# Copyright 2016 Julien Danjou
+# Copyright 2016–2021 Julien Danjou
 # Copyright 2016 Joshua Harlow
 # Copyright 2013-2014 Ray Holder
 #
@@ -16,74 +16,70 @@
 
 import abc
 import random
-
-import six
+import typing
 
 from tenacity import _utils
 
+if typing.TYPE_CHECKING:
+    from tenacity import RetryCallState
 
-@six.add_metaclass(abc.ABCMeta)
-class wait_base(object):
+
+class wait_base(abc.ABC):
     """Abstract base class for wait strategies."""
 
     @abc.abstractmethod
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
+    def __call__(self, retry_state: "RetryCallState") -> float:
         pass
 
-    def __add__(self, other):
+    def __add__(self, other: "wait_base") -> "wait_combine":
         return wait_combine(self, other)
 
-    def __radd__(self, other):
+    def __radd__(self, other: "wait_base") -> typing.Union["wait_combine", "wait_base"]:
         # make it possible to use multiple waits with the built-in sum function
-        if other == 0:
+        if other == 0:  # type: ignore[comparison-overlap]
             return self
         return self.__add__(other)
 
 
+WaitBaseT = typing.Union[wait_base, typing.Callable[["RetryCallState"], bool]]
+
+
 class wait_fixed(wait_base):
     """Wait strategy that waits a fixed amount of time between each retry."""
 
-    def __init__(self, wait):
-        self.wait_fixed = wait
+    def __init__(self, wait: _utils.time_unit_type) -> None:
+        self.wait_fixed = _utils.to_seconds(wait)
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
+    def __call__(self, retry_state: "RetryCallState") -> float:
         return self.wait_fixed
 
 
 class wait_none(wait_fixed):
     """Wait strategy that doesn't wait at all before retrying."""
 
-    def __init__(self):
-        super(wait_none, self).__init__(0)
+    def __init__(self) -> None:
+        super().__init__(0)
 
 
 class wait_random(wait_base):
     """Wait strategy that waits a random amount of time between min/max."""
 
-    def __init__(self, min=0, max=1):  # noqa
-        self.wait_random_min = min
-        self.wait_random_max = max
+    def __init__(self, min: _utils.time_unit_type = 0, max: _utils.time_unit_type = 1) -> None:  # noqa
+        self.wait_random_min = _utils.to_seconds(min)
+        self.wait_random_max = _utils.to_seconds(max)
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
-        return (self.wait_random_min +
-                (random.random() *
-                 (self.wait_random_max - self.wait_random_min)))
+    def __call__(self, retry_state: "RetryCallState") -> float:
+        return self.wait_random_min + (random.random() * (self.wait_random_max - self.wait_random_min))
 
 
 class wait_combine(wait_base):
     """Combine several waiting strategies."""
 
-    def __init__(self, *strategies):
+    def __init__(self, *strategies: wait_base) -> None:
         self.wait_funcs = strategies
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
-        return sum(map(
-            lambda x: x(previous_attempt_number, delay_since_first_attempt),
-            self.wait_funcs))
+    def __call__(self, retry_state: "RetryCallState") -> float:
+        return sum(x(retry_state=retry_state) for x in self.wait_funcs)
 
 
 class wait_chain(wait_base):
@@ -102,15 +98,13 @@ class wait_chain(wait_base):
                    thereafter.")
     """
 
-    def __init__(self, *strategies):
-        self.strategies = list(strategies)
+    def __init__(self, *strategies: wait_base) -> None:
+        self.strategies = strategies
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
-        wait_func = self.strategies[0]
-        if len(self.strategies) > 1:
-            self.strategies.pop(0)
-        return wait_func(previous_attempt_number, delay_since_first_attempt)
+    def __call__(self, retry_state: "RetryCallState") -> float:
+        wait_func_no = min(max(retry_state.attempt_number, 1), len(self.strategies))
+        wait_func = self.strategies[wait_func_no - 1]
+        return wait_func(retry_state=retry_state)
 
 
 class wait_incrementing(wait_base):
@@ -120,16 +114,18 @@ class wait_incrementing(wait_base):
     (and restricting the upper limit to some maximum value).
     """
 
-    def __init__(self, start=0, increment=100, max=_utils.MAX_WAIT):  # noqa
-        self.start = start
-        self.increment = increment
-        self.max = max
+    def __init__(
+        self,
+        start: _utils.time_unit_type = 0,
+        increment: _utils.time_unit_type = 100,
+        max: _utils.time_unit_type = _utils.MAX_WAIT,  # noqa
+    ) -> None:
+        self.start = _utils.to_seconds(start)
+        self.increment = _utils.to_seconds(increment)
+        self.max = _utils.to_seconds(max)
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
-        result = self.start + (
-            self.increment * (previous_attempt_number - 1)
-        )
+    def __call__(self, retry_state: "RetryCallState") -> float:
+        result = self.start + (self.increment * (retry_state.attempt_number - 1))
         return max(0, min(result, self.max))
 
 
@@ -137,7 +133,7 @@ class wait_exponential(wait_base):
     """Wait strategy that applies exponential backoff.
 
     It allows for a customized multiplier and an ability to restrict the
-    upper limit to some maximum value.
+    upper and lower limits to some maximum and minimum value.
 
     The intervals are fixed (i.e. there is no jitter), so this strategy is
     suitable for balancing retries against latency when a required resource is
@@ -146,31 +142,37 @@ class wait_exponential(wait_base):
     wait_random_exponential for the latter case.
     """
 
-    def __init__(self, multiplier=1, max=_utils.MAX_WAIT, exp_base=2):  # noqa
+    def __init__(
+        self,
+        multiplier: typing.Union[int, float] = 1,
+        max: _utils.time_unit_type = _utils.MAX_WAIT,  # noqa
+        exp_base: typing.Union[int, float] = 2,
+        min: _utils.time_unit_type = 0,  # noqa
+    ) -> None:
         self.multiplier = multiplier
-        self.max = max
+        self.min = _utils.to_seconds(min)
+        self.max = _utils.to_seconds(max)
         self.exp_base = exp_base
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
+    def __call__(self, retry_state: "RetryCallState") -> float:
         try:
-            exp = self.exp_base ** previous_attempt_number
+            exp = self.exp_base ** (retry_state.attempt_number - 1)
             result = self.multiplier * exp
         except OverflowError:
             return self.max
-        return max(0, min(result, self.max))
+        return max(max(0, self.min), min(result, self.max))
 
 
 class wait_random_exponential(wait_exponential):
     """Random wait with exponentially widening window.
 
     An exponential backoff strategy used to mediate contention between multiple
-    unco-ordinated processes for a shared resource in distributed systems. This
+    uncoordinated processes for a shared resource in distributed systems. This
     is the sense in which "exponential backoff" is meant in e.g. Ethernet
     networking, and corresponds to the "Full Jitter" algorithm described in
     this blog post:
 
-    https://www.awsarchitectureblog.com/2015/03/backoff.html
+    https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
 
     Each retry occurs at a random time in a geometrically expanding interval.
     It allows for a custom multiplier and an ability to restrict the upper
@@ -184,10 +186,43 @@ class wait_random_exponential(wait_exponential):
     When waiting for an unavailable resource to become available again, as
     opposed to trying to resolve contention for a shared resource, the
     wait_exponential strategy (which uses a fixed interval) may be preferable.
+
     """
 
-    def __call__(self, previous_attempt_number, delay_since_first_attempt,
-                 last_result=None):
-        high = super(wait_random_exponential, self).__call__(
-            previous_attempt_number, delay_since_first_attempt)
+    def __call__(self, retry_state: "RetryCallState") -> float:
+        high = super().__call__(retry_state=retry_state)
         return random.uniform(0, high)
+
+
+class wait_exponential_jitter(wait_base):
+    """Wait strategy that applies exponential backoff and jitter.
+
+    It allows for a customized initial wait, maximum wait and jitter.
+
+    This implements the strategy described here:
+    https://cloud.google.com/storage/docs/retry-strategy
+
+    The wait time is min(initial * 2**n + random.uniform(0, jitter), maximum)
+    where n is the retry count.
+    """
+
+    def __init__(
+        self,
+        initial: float = 1,
+        max: float = _utils.MAX_WAIT,  # noqa
+        exp_base: float = 2,
+        jitter: float = 1,
+    ) -> None:
+        self.initial = initial
+        self.max = max
+        self.exp_base = exp_base
+        self.jitter = jitter
+
+    def __call__(self, retry_state: "RetryCallState") -> float:
+        jitter = random.uniform(0, self.jitter)
+        try:
+            exp = self.exp_base ** (retry_state.attempt_number - 1)
+            result = self.initial * exp + jitter
+        except OverflowError:
+            result = self.max
+        return max(0, min(result, self.max))
